#define _CFFI_

/* We try to define Py_LIMITED_API before including Python.h.

   Mess: we can only define it if Py_DEBUG, Py_TRACE_REFS and
   Py_REF_DEBUG are not defined.  This is a best-effort approximation:
   we can learn about Py_DEBUG from pyconfig.h, but it is unclear if
   the same works for the other two macros.  Py_DEBUG implies them,
   but not the other way around.

   Issue #350 is still open: on Windows, the code here causes it to link
   with PYTHON36.DLL (for example) instead of PYTHON3.DLL.  A fix was
   attempted in 164e526a5515 and 14ce6985e1c3, but reverted: virtualenv
   does not make PYTHON3.DLL available, and so the "correctly" compiled
   version would not run inside a virtualenv.  We will re-apply the fix
   after virtualenv has been fixed for some time.  For explanation, see
   issue #355.  For a workaround if you want PYTHON3.DLL and don't worry
   about virtualenv, see issue #350.  See also 'py_limited_api' in
   setuptools_ext.py.
*/
#if !defined(_CFFI_USE_EMBEDDING) && !defined(Py_LIMITED_API)
#  include <pyconfig.h>
#  if !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG)
#    define Py_LIMITED_API
#  endif
#endif

#include <Python.h>
#ifdef __cplusplus
extern "C" {
#endif
#include <stddef.h>

/* This part is from file 'cffi/parse_c_type.h'.  It is copied at the
   beginning of C sources generated by CFFI's ffi.set_source(). */

typedef void *_cffi_opcode_t;

#define _CFFI_OP(opcode, arg)   (_cffi_opcode_t)(opcode | (((uintptr_t)(arg)) << 8))
#define _CFFI_GETOP(cffi_opcode)    ((unsigned char)(uintptr_t)cffi_opcode)
#define _CFFI_GETARG(cffi_opcode)   (((intptr_t)cffi_opcode) >> 8)

#define _CFFI_OP_PRIMITIVE       1
#define _CFFI_OP_POINTER         3
#define _CFFI_OP_ARRAY           5
#define _CFFI_OP_OPEN_ARRAY      7
#define _CFFI_OP_STRUCT_UNION    9
#define _CFFI_OP_ENUM           11
#define _CFFI_OP_FUNCTION       13
#define _CFFI_OP_FUNCTION_END   15
#define _CFFI_OP_NOOP           17
#define _CFFI_OP_BITFIELD       19
#define _CFFI_OP_TYPENAME       21
#define _CFFI_OP_CPYTHON_BLTN_V 23   // varargs
#define _CFFI_OP_CPYTHON_BLTN_N 25   // noargs
#define _CFFI_OP_CPYTHON_BLTN_O 27   // O  (i.e. a single arg)
#define _CFFI_OP_CONSTANT       29
#define _CFFI_OP_CONSTANT_INT   31
#define _CFFI_OP_GLOBAL_VAR     33
#define _CFFI_OP_DLOPEN_FUNC    35
#define _CFFI_OP_DLOPEN_CONST   37
#define _CFFI_OP_GLOBAL_VAR_F   39
#define _CFFI_OP_EXTERN_PYTHON  41

#define _CFFI_PRIM_VOID          0
#define _CFFI_PRIM_BOOL          1
#define _CFFI_PRIM_CHAR          2
#define _CFFI_PRIM_SCHAR         3
#define _CFFI_PRIM_UCHAR         4
#define _CFFI_PRIM_SHORT         5
#define _CFFI_PRIM_USHORT        6
#define _CFFI_PRIM_INT           7
#define _CFFI_PRIM_UINT          8
#define _CFFI_PRIM_LONG          9
#define _CFFI_PRIM_ULONG        10
#define _CFFI_PRIM_LONGLONG     11
#define _CFFI_PRIM_ULONGLONG    12
#define _CFFI_PRIM_FLOAT        13
#define _CFFI_PRIM_DOUBLE       14
#define _CFFI_PRIM_LONGDOUBLE   15

#define _CFFI_PRIM_WCHAR        16
#define _CFFI_PRIM_INT8         17
#define _CFFI_PRIM_UINT8        18
#define _CFFI_PRIM_INT16        19
#define _CFFI_PRIM_UINT16       20
#define _CFFI_PRIM_INT32        21
#define _CFFI_PRIM_UINT32       22
#define _CFFI_PRIM_INT64        23
#define _CFFI_PRIM_UINT64       24
#define _CFFI_PRIM_INTPTR       25
#define _CFFI_PRIM_UINTPTR      26
#define _CFFI_PRIM_PTRDIFF      27
#define _CFFI_PRIM_SIZE         28
#define _CFFI_PRIM_SSIZE        29
#define _CFFI_PRIM_INT_LEAST8   30
#define _CFFI_PRIM_UINT_LEAST8  31
#define _CFFI_PRIM_INT_LEAST16  32
#define _CFFI_PRIM_UINT_LEAST16 33
#define _CFFI_PRIM_INT_LEAST32  34
#define _CFFI_PRIM_UINT_LEAST32 35
#define _CFFI_PRIM_INT_LEAST64  36
#define _CFFI_PRIM_UINT_LEAST64 37
#define _CFFI_PRIM_INT_FAST8    38
#define _CFFI_PRIM_UINT_FAST8   39
#define _CFFI_PRIM_INT_FAST16   40
#define _CFFI_PRIM_UINT_FAST16  41
#define _CFFI_PRIM_INT_FAST32   42
#define _CFFI_PRIM_UINT_FAST32  43
#define _CFFI_PRIM_INT_FAST64   44
#define _CFFI_PRIM_UINT_FAST64  45
#define _CFFI_PRIM_INTMAX       46
#define _CFFI_PRIM_UINTMAX      47
#define _CFFI_PRIM_FLOATCOMPLEX 48
#define _CFFI_PRIM_DOUBLECOMPLEX 49
#define _CFFI_PRIM_CHAR16       50
#define _CFFI_PRIM_CHAR32       51

#define _CFFI__NUM_PRIM         52
#define _CFFI__UNKNOWN_PRIM           (-1)
#define _CFFI__UNKNOWN_FLOAT_PRIM     (-2)
#define _CFFI__UNKNOWN_LONG_DOUBLE    (-3)

#define _CFFI__IO_FILE_STRUCT         (-1)


struct _cffi_global_s {
    const char *name;
    void *address;
    _cffi_opcode_t type_op;
    void *size_or_direct_fn;  // OP_GLOBAL_VAR: size, or 0 if unknown
                              // OP_CPYTHON_BLTN_*: addr of direct function
};

struct _cffi_getconst_s {
    unsigned long long value;
    const struct _cffi_type_context_s *ctx;
    int gindex;
};

struct _cffi_struct_union_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_STRUCT_UNION
    int flags;               // _CFFI_F_* flags below
    size_t size;
    int alignment;
    int first_field_index;   // -> _cffi_fields array
    int num_fields;
};
#define _CFFI_F_UNION         0x01   // is a union, not a struct
#define _CFFI_F_CHECK_FIELDS  0x02   // complain if fields are not in the
                                     // "standard layout" or if some are missing
#define _CFFI_F_PACKED        0x04   // for CHECK_FIELDS, assume a packed struct
#define _CFFI_F_EXTERNAL      0x08   // in some other ffi.include()
#define _CFFI_F_OPAQUE        0x10   // opaque

struct _cffi_field_s {
    const char *name;
    size_t field_offset;
    size_t field_size;
    _cffi_opcode_t field_type_op;
};

struct _cffi_enum_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_ENUM
    int type_prim;           // _CFFI_PRIM_xxx
    const char *enumerators; // comma-delimited string
};

struct _cffi_typename_s {
    const char *name;
    int type_index;   /* if opaque, points to a possibly artificial
                         OP_STRUCT which is itself opaque */
};

struct _cffi_type_context_s {
    _cffi_opcode_t *types;
    const struct _cffi_global_s *globals;
    const struct _cffi_field_s *fields;
    const struct _cffi_struct_union_s *struct_unions;
    const struct _cffi_enum_s *enums;
    const struct _cffi_typename_s *typenames;
    int num_globals;
    int num_struct_unions;
    int num_enums;
    int num_typenames;
    const char *const *includes;
    int num_types;
    int flags;      /* future extension */
};

struct _cffi_parse_info_s {
    const struct _cffi_type_context_s *ctx;
    _cffi_opcode_t *output;
    unsigned int output_size;
    size_t error_location;
    const char *error_message;
};

struct _cffi_externpy_s {
    const char *name;
    size_t size_of_result;
    void *reserved1, *reserved2;
};

#ifdef _CFFI_INTERNAL
static int parse_c_type(struct _cffi_parse_info_s *info, const char *input);
static int search_in_globals(const struct _cffi_type_context_s *ctx,
                             const char *search, size_t search_len);
static int search_in_struct_unions(const struct _cffi_type_context_s *ctx,
                                   const char *search, size_t search_len);
#endif

/* this block of #ifs should be kept exactly identical between
   c/_cffi_backend.c, cffi/vengine_cpy.py, cffi/vengine_gen.py
   and cffi/_cffi_include.h */
#if defined(_MSC_VER)
# include <malloc.h>   /* for alloca() */
# if _MSC_VER < 1600   /* MSVC < 2010 */
   typedef __int8 int8_t;
   typedef __int16 int16_t;
   typedef __int32 int32_t;
   typedef __int64 int64_t;
   typedef unsigned __int8 uint8_t;
   typedef unsigned __int16 uint16_t;
   typedef unsigned __int32 uint32_t;
   typedef unsigned __int64 uint64_t;
   typedef __int8 int_least8_t;
   typedef __int16 int_least16_t;
   typedef __int32 int_least32_t;
   typedef __int64 int_least64_t;
   typedef unsigned __int8 uint_least8_t;
   typedef unsigned __int16 uint_least16_t;
   typedef unsigned __int32 uint_least32_t;
   typedef unsigned __int64 uint_least64_t;
   typedef __int8 int_fast8_t;
   typedef __int16 int_fast16_t;
   typedef __int32 int_fast32_t;
   typedef __int64 int_fast64_t;
   typedef unsigned __int8 uint_fast8_t;
   typedef unsigned __int16 uint_fast16_t;
   typedef unsigned __int32 uint_fast32_t;
   typedef unsigned __int64 uint_fast64_t;
   typedef __int64 intmax_t;
   typedef unsigned __int64 uintmax_t;
# else
#  include <stdint.h>
# endif
# if _MSC_VER < 1800   /* MSVC < 2013 */
#  ifndef __cplusplus
    typedef unsigned char _Bool;
#  endif
# endif
#else
# include <stdint.h>
# if (defined (__SVR4) && defined (__sun)) || defined(_AIX) || defined(__hpux)
#  include <alloca.h>
# endif
#endif

#ifdef __GNUC__
# define _CFFI_UNUSED_FN  __attribute__((unused))
#else
# define _CFFI_UNUSED_FN  /* nothing */
#endif

#ifdef __cplusplus
# ifndef _Bool
   typedef bool _Bool;   /* semi-hackish: C++ has no _Bool; bool is builtin */
# endif
#endif

/**********  CPython-specific section  **********/
#ifndef PYPY_VERSION


#if PY_MAJOR_VERSION >= 3
# define PyInt_FromLong PyLong_FromLong
#endif

#define _cffi_from_c_double PyFloat_FromDouble
#define _cffi_from_c_float PyFloat_FromDouble
#define _cffi_from_c_long PyInt_FromLong
#define _cffi_from_c_ulong PyLong_FromUnsignedLong
#define _cffi_from_c_longlong PyLong_FromLongLong
#define _cffi_from_c_ulonglong PyLong_FromUnsignedLongLong
#define _cffi_from_c__Bool PyBool_FromLong

#define _cffi_to_c_double PyFloat_AsDouble
#define _cffi_to_c_float PyFloat_AsDouble

#define _cffi_from_c_int(x, type)                                        \
    (((type)-1) > 0 ? /* unsigned */                                     \
        (sizeof(type) < sizeof(long) ?                                   \
            PyInt_FromLong((long)x) :                                    \
         sizeof(type) == sizeof(long) ?                                  \
            PyLong_FromUnsignedLong((unsigned long)x) :                  \
            PyLong_FromUnsignedLongLong((unsigned long long)x)) :        \
        (sizeof(type) <= sizeof(long) ?                                  \
            PyInt_FromLong((long)x) :                                    \
            PyLong_FromLongLong((long long)x)))

#define _cffi_to_c_int(o, type)                                          \
    ((type)(                                                             \
     sizeof(type) == 1 ? (((type)-1) > 0 ? (type)_cffi_to_c_u8(o)        \
                                         : (type)_cffi_to_c_i8(o)) :     \
     sizeof(type) == 2 ? (((type)-1) > 0 ? (type)_cffi_to_c_u16(o)       \
                                         : (type)_cffi_to_c_i16(o)) :    \
     sizeof(type) == 4 ? (((type)-1) > 0 ? (type)_cffi_to_c_u32(o)       \
                                         : (type)_cffi_to_c_i32(o)) :    \
     sizeof(type) == 8 ? (((type)-1) > 0 ? (type)_cffi_to_c_u64(o)       \
                                         : (type)_cffi_to_c_i64(o)) :    \
     (Py_FatalError("unsupported size for type " #type), (type)0)))

#define _cffi_to_c_i8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[1])
#define _cffi_to_c_u8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[2])
#define _cffi_to_c_i16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[3])
#define _cffi_to_c_u16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[4])
#define _cffi_to_c_i32                                                   \
                 ((int(*)(PyObject *))_cffi_exports[5])
#define _cffi_to_c_u32                                                   \
                 ((unsigned int(*)(PyObject *))_cffi_exports[6])
#define _cffi_to_c_i64                                                   \
                 ((long long(*)(PyObject *))_cffi_exports[7])
#define _cffi_to_c_u64                                                   \
                 ((unsigned long long(*)(PyObject *))_cffi_exports[8])
#define _cffi_to_c_char                                                  \
                 ((int(*)(PyObject *))_cffi_exports[9])
#define _cffi_from_c_pointer                                             \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[10])
#define _cffi_to_c_pointer                                               \
    ((char *(*)(PyObject *, struct _cffi_ctypedescr *))_cffi_exports[11])
#define _cffi_get_struct_layout                                          \
    not used any more
#define _cffi_restore_errno                                              \
    ((void(*)(void))_cffi_exports[13])
#define _cffi_save_errno                                                 \
    ((void(*)(void))_cffi_exports[14])
#define _cffi_from_c_char                                                \
    ((PyObject *(*)(char))_cffi_exports[15])
#define _cffi_from_c_deref                                               \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[16])
#define _cffi_to_c                                                       \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[17])
#define _cffi_from_c_struct                                              \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[18])
#define _cffi_to_c_wchar_t                                               \
    ((_cffi_wchar_t(*)(PyObject *))_cffi_exports[19])
#define _cffi_from_c_wchar_t                                             \
    ((PyObject *(*)(_cffi_wchar_t))_cffi_exports[20])
#define _cffi_to_c_long_double                                           \
    ((long double(*)(PyObject *))_cffi_exports[21])
#define _cffi_to_c__Bool                                                 \
    ((_Bool(*)(PyObject *))_cffi_exports[22])
#define _cffi_prepare_pointer_call_argument                              \
    ((Py_ssize_t(*)(struct _cffi_ctypedescr *,                           \
                    PyObject *, char **))_cffi_exports[23])
#define _cffi_convert_array_from_object                                  \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[24])
#define _CFFI_CPIDX  25
#define _cffi_call_python                                                \
    ((void(*)(struct _cffi_externpy_s *, char *))_cffi_exports[_CFFI_CPIDX])
#define _cffi_to_c_wchar3216_t                                           \
    ((int(*)(PyObject *))_cffi_exports[26])
#define _cffi_from_c_wchar3216_t                                         \
    ((PyObject *(*)(int))_cffi_exports[27])
#define _CFFI_NUM_EXPORTS 28

struct _cffi_ctypedescr;

static void *_cffi_exports[_CFFI_NUM_EXPORTS];

#define _cffi_type(index)   (                           \
    assert((((uintptr_t)_cffi_types[index]) & 1) == 0), \
    (struct _cffi_ctypedescr *)_cffi_types[index])

static PyObject *_cffi_init(const char *module_name, Py_ssize_t version,
                            const struct _cffi_type_context_s *ctx)
{
    PyObject *module, *o_arg, *new_module;
    void *raw[] = {
        (void *)module_name,
        (void *)version,
        (void *)_cffi_exports,
        (void *)ctx,
    };

    module = PyImport_ImportModule("_cffi_backend");
    if (module == NULL)
        goto failure;

    o_arg = PyLong_FromVoidPtr((void *)raw);
    if (o_arg == NULL)
        goto failure;

    new_module = PyObject_CallMethod(
        module, (char *)"_init_cffi_1_0_external_module", (char *)"O", o_arg);

    Py_DECREF(o_arg);
    Py_DECREF(module);
    return new_module;

  failure:
    Py_XDECREF(module);
    return NULL;
}


#ifdef HAVE_WCHAR_H
typedef wchar_t _cffi_wchar_t;
#else
typedef uint16_t _cffi_wchar_t;   /* same random pick as _cffi_backend.c */
#endif

_CFFI_UNUSED_FN static uint16_t _cffi_to_c_char16_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return (uint16_t)_cffi_to_c_wchar_t(o);
    else
        return (uint16_t)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char16_t(uint16_t x)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t((int)x);
}

_CFFI_UNUSED_FN static int _cffi_to_c_char32_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return (int)_cffi_to_c_wchar_t(o);
    else
        return (int)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char32_t(int x)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t(x);
}


/**********  end CPython-specific section  **********/
#else
_CFFI_UNUSED_FN
static void (*_cffi_call_python_org)(struct _cffi_externpy_s *, char *);
# define _cffi_call_python  _cffi_call_python_org
#endif


#define _cffi_array_len(array)   (sizeof(array) / sizeof((array)[0]))

#define _cffi_prim_int(size, sign)                                      \
    ((size) == 1 ? ((sign) ? _CFFI_PRIM_INT8  : _CFFI_PRIM_UINT8)  :    \
     (size) == 2 ? ((sign) ? _CFFI_PRIM_INT16 : _CFFI_PRIM_UINT16) :    \
     (size) == 4 ? ((sign) ? _CFFI_PRIM_INT32 : _CFFI_PRIM_UINT32) :    \
     (size) == 8 ? ((sign) ? _CFFI_PRIM_INT64 : _CFFI_PRIM_UINT64) :    \
     _CFFI__UNKNOWN_PRIM)

#define _cffi_prim_float(size)                                          \
    ((size) == sizeof(float) ? _CFFI_PRIM_FLOAT :                       \
     (size) == sizeof(double) ? _CFFI_PRIM_DOUBLE :                     \
     (size) == sizeof(long double) ? _CFFI__UNKNOWN_LONG_DOUBLE :       \
     _CFFI__UNKNOWN_FLOAT_PRIM)

#define _cffi_check_int(got, got_nonpos, expected)      \
    ((got_nonpos) == (expected <= 0) &&                 \
     (got) == (unsigned long long)expected)

#ifdef MS_WIN32
# define _cffi_stdcall  __stdcall
#else
# define _cffi_stdcall  /* nothing */
#endif

#ifdef __cplusplus
}
#endif

/************************************************************/

#include <portaudio.h>
#include <pa_ringbuffer.h>
#include "py_pastream.h"
#ifdef PYPA_DEBUG
#include <stdio.h>
#endif


void init_stream(Py_PaStream *stream)
{
    *stream = Py_PaStream_default;
    reset_stream(stream);
};

void reset_stream(Py_PaStream *stream) {
    memset((void *) &stream->lastTime, 0, sizeof(PaStreamCallbackTimeInfo));

    stream->last_callback = paContinue;
    stream->status = 0;
    stream->frame_count = 0;
    stream->errorMsg[0] = '\0';
    stream->xruns = 0;
    stream->inputUnderflows = 0;
    stream->inputOverflows = 0;
    stream->outputUnderflows = 0;
    stream->outputOverflows = 0;
    if ( stream->_autoframes && stream->pad >= 0 )
        stream->frames = -1;
    stream->_autoframes = 0;
};

int callback(
    void* in_data,
    void* out_data,
    unsigned long frame_count,
    const PaStreamCallbackTimeInfo* timeInfo,
    PaStreamCallbackFlags status,
    void *user_data)
{
    unsigned long offset = 0;
    ring_buffer_size_t frames_left = frame_count;
    ring_buffer_size_t oframes_left = frame_count;
    ring_buffer_size_t oframes, iframes, tempframes;
    Py_PaStream * stream = (Py_PaStream *) user_data;
    long long frames = stream->frames;
    long pad = stream->pad;
    int i, j, k;
    unsigned char temp, *ptemp;
    unsigned char samplesize;

    if ( status & 0xF ) {
        if ( status & paInputUnderflow )
            stream->inputUnderflows++;
        if ( status & paInputOverflow )
            stream->inputOverflows++;
        if ( status & paOutputUnderflow )
            stream->outputUnderflows++;
        if ( status & paOutputOverflow )
            stream->outputOverflows++;
        stream->xruns++;
        stream->status |= status;
    }

    if ( frames >= 0 ) {
        // exit point (1 of 2)
        // We've surpassed frames: this is our last callback
        if ( stream->frame_count + frames_left >= frames ) {
            frames_left = frames - stream->frame_count;
            stream->last_callback = paComplete;
        }

        // Calcuate how many output frames are left to read (minus any padding)
        tempframes = frames - pad - stream->frame_count;
        if ( pad >= 0 && frames_left >= tempframes ) {
            oframes_left = tempframes > 0 ? tempframes : 0;
        }
    }

    if ( out_data != NULL && (stream->txbuffer == NULL || stream->_autoframes) ) {
        memset((unsigned char *) out_data, 0, frame_count*stream->txElementSize);
    }
    else if ( stream->txbuffer != NULL ) {
        oframes = PaUtil_ReadRingBuffer(stream->txbuffer, out_data, oframes_left);

        // loop mode: this assumes that the buffer is only written before
        // stream is started (and never while its active), so we can safely
        // rewind when we hit the end
        while ( oframes < oframes_left && stream->loop ) {
            tempframes = stream->txbuffer->readIndex;

            // just in case frames == 0 or by some mistake the txbuffer is empty
            if ( tempframes == 0 ) break;

            PaUtil_FlushRingBuffer(stream->txbuffer);
            PaUtil_AdvanceRingBufferWriteIndex(stream->txbuffer, tempframes);
            oframes += PaUtil_ReadRingBuffer(stream->txbuffer,
                           (unsigned char *) out_data + oframes*stream->txElementSize,
                           oframes_left - oframes);
        }

        // Re-map data to requested output channels
        if ( stream->txmapping != NULL ) {
            samplesize = stream->txElementSize / stream->txchannels;
            memcpy(stream->_mapping, stream->txmapping, stream->txchannels);

            for ( i = 0 ; i < stream->txchannels ; i++ ) {
                // inchannel = stream->_mapping[i];
                // outchannel = i + 1;
                if ( stream->_mapping[i] == i + 1 || stream->_mapping[i] == 0 )
                    continue;

                ptemp = (unsigned char *) out_data;
                if ( i + 1 == stream->txchannels || stream->_mapping[stream->_mapping[i] - 1] == 0 ) {
                    for ( j = 0 ; j < oframes ; j++, ptemp += stream->txElementSize ) {
                        memcpy(ptemp + i * samplesize, ptemp + (stream->_mapping[i] - 1)*samplesize, samplesize);
                    }
                }
                else {
                    for ( j = 0 ; j < oframes ; j++, ptemp += stream->txElementSize ) {
                        for ( k = 0 ; k < samplesize ; k++) {
                            temp = ptemp[i * samplesize + k];
                            ptemp[i * samplesize + k] = ptemp[(stream->_mapping[i] - 1)*samplesize + k];
                            ptemp[(stream->_mapping[i] - 1)*samplesize + k] = temp;
                        }
                    }
                }

                for ( k = i + 1 ; k < stream->txchannels ; k++) {
                    if (stream->_mapping[k] == i + 1)
                        stream->_mapping[k] = stream->_mapping[i];
                    else if (stream->_mapping[k] == stream->_mapping[i])
                        stream->_mapping[k] = i + 1;
                }
            }
            // Do channel zero out as a last step to avoid unnecessary swapping
            for ( i = 0 ; i < stream->txchannels ; i++ ) {
                if ( stream->_mapping[i] != 0 )
                    continue;

                ptemp = (unsigned char *) out_data + i * samplesize;
                for ( j = 0 ; j < oframes ; j++, ptemp += stream->txElementSize ) {
                    memset(ptemp, 0, samplesize);
                }
            }
        }

        // No matter what happens, fill the remainder of the output buffer with zeros
        if ( oframes < frame_count )
            memset((unsigned char *) out_data + oframes*stream->txElementSize,
                   0,
                   (frame_count - oframes)*stream->txElementSize);

        if ( oframes < frames_left) {
            if ( frames < 0 ) {
                // Now that we've reached the end of buffer, calculate our
                // final frame count including padding and enter autoframes
                // mode
                if ( pad >= 0 ) {
                    stream->_autoframes = 1;
                    frames = stream->frames = stream->frame_count + oframes + pad;

                    // exit point (2 of 2)
                    // We don't want to do an unncessary callback; end here
                    if ( stream->frame_count + frames_left >= frames ) {
                        if ( stream->frame_count <= frames )
                            frames_left = frames - stream->frame_count;
                        else
                            frames_left = 0;
                        stream->last_callback = paComplete;
                    }
                }
            }
            else if ( pad < 0 ) { // Just pad out to stream->frames samples
                stream->_autoframes = 1;
            }
            else if ( oframes < oframes_left ) {
                strcpy(stream->errorMsg, "BufferEmpty");
                stream->frame_count += oframes;
                return stream->last_callback = paAbort;
            }
        }
    }

#ifdef PYPA_WIREMODE
    if ( out_data != NULL && in_data != NULL ) {
        memcpy(in_data, out_data, frame_count * stream->txElementSize);
    }
#endif

    if ( stream->rxbuffer != NULL && stream->frame_count + frames_left > stream->offset ) {
        if ( stream->frame_count < stream->offset ) {
            offset = stream->offset - stream->frame_count;
            frames_left -= offset;
            in_data = (unsigned char *) in_data + offset*stream->rxElementSize;
        }

        if ( stream->rxmapping != NULL ) {
            samplesize = stream->rxElementSize / stream->rxchannels;
            memcpy(stream->_mapping, stream->rxmapping, stream->rxchannels);

            for ( i = 0 ; i < stream->rxchannels ; i++ ) {
                // inchannel = stream->_mapping[i];
                // outchannel = i + 1;
                if ( stream->_mapping[i] == i + 1 || stream->_mapping[i] == 0 )
                    continue;

                ptemp = (unsigned char *) in_data;
                if ( i + 1 == stream->rxchannels || stream->_mapping[stream->_mapping[i] - 1] == 0 ) {
                    for ( j = 0 ; j < frames_left ; j++, ptemp += stream->rxElementSize ) {
                        memcpy(ptemp + i * samplesize, ptemp + (stream->_mapping[i] - 1)*samplesize, samplesize);
                    }
                }
                else {
                    for ( j = 0 ; j < frames_left ; j++, ptemp += stream->rxElementSize ) {
                        for ( k = 0 ; k < samplesize ; k++) {
                            temp = ptemp[i * samplesize + k];
                            ptemp[i * samplesize + k] = ptemp[(stream->_mapping[i] - 1)*samplesize + k];
                            ptemp[(stream->_mapping[i] - 1)*samplesize + k] = temp;
                        }
                    }
                }

                for ( k = i + 1 ; k < stream->rxchannels ; k++) {
                    if (stream->_mapping[k] == i + 1)
                        stream->_mapping[k] = stream->_mapping[i];
                    else if (stream->_mapping[k] == stream->_mapping[i])
                        stream->_mapping[k] = i + 1;
                }
            }
            // Do channel zero out as a last step to avoid unnecessary swapping
            for ( i = 0 ; i < stream->rxchannels ; i++ ) {
                if ( stream->_mapping[i] != 0 )
                    continue;

                ptemp = (unsigned char *) in_data + i * samplesize;
                for ( j = 0 ; j < frames_left ; j++, ptemp += stream->rxElementSize ) {
                    memset(ptemp, 0, samplesize);
                }
            }
        }

        iframes = PaUtil_WriteRingBuffer(stream->rxbuffer, (const void *) in_data, frames_left);
        if ( iframes < frames_left ) {
            strcpy(stream->errorMsg, "BufferFull");
            stream->frame_count += iframes;
            return stream->last_callback = paAbort;
        }
    }

    stream->lastTime = *timeInfo;
    stream->frame_count += frame_count;
    return stream->last_callback;
}


/************************************************************/

static void *_cffi_types[] = {
/*  0 */ _CFFI_OP(_CFFI_OP_FUNCTION, 53), // int()(void *, void *, unsigned long, PaStreamCallbackTimeInfo const *, unsigned long, void *)
/*  1 */ _CFFI_OP(_CFFI_OP_POINTER, 58), // void *
/*  2 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/*  3 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 10), // unsigned long
/*  4 */ _CFFI_OP(_CFFI_OP_POINTER, 45), // PaStreamCallbackTimeInfo const *
/*  5 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 10),
/*  6 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/*  7 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/*  8 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, long)
/*  9 */ _CFFI_OP(_CFFI_OP_POINTER, 46), // PaUtilRingBuffer *
/* 10 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9), // long
/* 11 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 12 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, long, long, void *)
/* 13 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 14 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 15 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 16 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/* 17 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 18 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, long, void * *, long *, void * *, long *)
/* 19 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 20 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 21 */ _CFFI_OP(_CFFI_OP_POINTER, 1), // void * *
/* 22 */ _CFFI_OP(_CFFI_OP_POINTER, 10), // long *
/* 23 */ _CFFI_OP(_CFFI_OP_NOOP, 21),
/* 24 */ _CFFI_OP(_CFFI_OP_NOOP, 22),
/* 25 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 26 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, void *, long)
/* 27 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 28 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/* 29 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 30 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 31 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, void const *, long)
/* 32 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 33 */ _CFFI_OP(_CFFI_OP_POINTER, 58), // void const *
/* 34 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 35 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 36 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer const *)
/* 37 */ _CFFI_OP(_CFFI_OP_POINTER, 46), // PaUtilRingBuffer const *
/* 38 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 39 */ _CFFI_OP(_CFFI_OP_FUNCTION, 58), // void()(PaUtilRingBuffer *)
/* 40 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 41 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 42 */ _CFFI_OP(_CFFI_OP_FUNCTION, 58), // void()(Py_PaStream *)
/* 43 */ _CFFI_OP(_CFFI_OP_POINTER, 47), // Py_PaStream *
/* 44 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 45 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 0), // PaStreamCallbackTimeInfo
/* 46 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 1), // PaUtilRingBuffer
/* 47 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 2), // Py_PaStream
/* 48 */ _CFFI_OP(_CFFI_OP_POINTER, 49), // char *
/* 49 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 2), // char
/* 50 */ _CFFI_OP(_CFFI_OP_ARRAY, 49), // char[128]
/* 51 */ (_cffi_opcode_t)(128),
/* 52 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14), // double
/* 53 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 7), // int
/* 54 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 11), // long long
/* 55 */ _CFFI_OP(_CFFI_OP_POINTER, 56), // unsigned char *
/* 56 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 4), // unsigned char
/* 57 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 12), // unsigned long long
/* 58 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 0), // void
};

static void _cffi_const_Py_PaStream_default(char *o)
{
  *(Py_PaStream *)o = Py_PaStream_default;
}

static long _cffi_d_PaUtil_AdvanceRingBufferReadIndex(PaUtilRingBuffer * x0, long x1)
{
  return PaUtil_AdvanceRingBufferReadIndex(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_AdvanceRingBufferReadIndex(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  Py_ssize_t datasize;
  long result;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "PaUtil_AdvanceRingBufferReadIndex", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(9), arg0) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_AdvanceRingBufferReadIndex(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, long);
}
#else
#  define _cffi_f_PaUtil_AdvanceRingBufferReadIndex _cffi_d_PaUtil_AdvanceRingBufferReadIndex
#endif

static long _cffi_d_PaUtil_AdvanceRingBufferWriteIndex(PaUtilRingBuffer * x0, long x1)
{
  return PaUtil_AdvanceRingBufferWriteIndex(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_AdvanceRingBufferWriteIndex(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  Py_ssize_t datasize;
  long result;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "PaUtil_AdvanceRingBufferWriteIndex", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(9), arg0) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_AdvanceRingBufferWriteIndex(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, long);
}
#else
#  define _cffi_f_PaUtil_AdvanceRingBufferWriteIndex _cffi_d_PaUtil_AdvanceRingBufferWriteIndex
#endif

static void _cffi_d_PaUtil_FlushRingBuffer(PaUtilRingBuffer * x0)
{
  PaUtil_FlushRingBuffer(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_FlushRingBuffer(PyObject *self, PyObject *arg0)
{
  PaUtilRingBuffer * x0;
  Py_ssize_t datasize;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(9), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { PaUtil_FlushRingBuffer(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_PaUtil_FlushRingBuffer _cffi_d_PaUtil_FlushRingBuffer
#endif

static long _cffi_d_PaUtil_GetRingBufferReadAvailable(PaUtilRingBuffer const * x0)
{
  return PaUtil_GetRingBufferReadAvailable(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_GetRingBufferReadAvailable(PyObject *self, PyObject *arg0)
{
  PaUtilRingBuffer const * x0;
  Py_ssize_t datasize;
  long result;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(37), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer const *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(37), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_GetRingBufferReadAvailable(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, long);
}
#else
#  define _cffi_f_PaUtil_GetRingBufferReadAvailable _cffi_d_PaUtil_GetRingBufferReadAvailable
#endif

static long _cffi_d_PaUtil_GetRingBufferReadRegions(PaUtilRingBuffer * x0, long x1, void * * x2, long * x3, void * * x4, long * x5)
{
  return PaUtil_GetRingBufferReadRegions(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_GetRingBufferReadRegions(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  void * * x2;
  long * x3;
  void * * x4;
  long * x5;
  Py_ssize_t datasize;
  long result;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "PaUtil_GetRingBufferReadRegions", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(9), arg0) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(21), arg2, (char **)&x2);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x2 = (void * *)alloca((size_t)datasize);
    memset((void *)x2, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x2, _cffi_type(21), arg2) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(22), arg3, (char **)&x3);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x3 = (long *)alloca((size_t)datasize);
    memset((void *)x3, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x3, _cffi_type(22), arg3) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(21), arg4, (char **)&x4);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x4 = (void * *)alloca((size_t)datasize);
    memset((void *)x4, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x4, _cffi_type(21), arg4) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(22), arg5, (char **)&x5);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x5 = (long *)alloca((size_t)datasize);
    memset((void *)x5, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x5, _cffi_type(22), arg5) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_GetRingBufferReadRegions(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, long);
}
#else
#  define _cffi_f_PaUtil_GetRingBufferReadRegions _cffi_d_PaUtil_GetRingBufferReadRegions
#endif

static long _cffi_d_PaUtil_GetRingBufferWriteAvailable(PaUtilRingBuffer const * x0)
{
  return PaUtil_GetRingBufferWriteAvailable(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_GetRingBufferWriteAvailable(PyObject *self, PyObject *arg0)
{
  PaUtilRingBuffer const * x0;
  Py_ssize_t datasize;
  long result;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(37), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer const *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(37), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_GetRingBufferWriteAvailable(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, long);
}
#else
#  define _cffi_f_PaUtil_GetRingBufferWriteAvailable _cffi_d_PaUtil_GetRingBufferWriteAvailable
#endif

static long _cffi_d_PaUtil_GetRingBufferWriteRegions(PaUtilRingBuffer * x0, long x1, void * * x2, long * x3, void * * x4, long * x5)
{
  return PaUtil_GetRingBufferWriteRegions(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_GetRingBufferWriteRegions(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  void * * x2;
  long * x3;
  void * * x4;
  long * x5;
  Py_ssize_t datasize;
  long result;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "PaUtil_GetRingBufferWriteRegions", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(9), arg0) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(21), arg2, (char **)&x2);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x2 = (void * *)alloca((size_t)datasize);
    memset((void *)x2, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x2, _cffi_type(21), arg2) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(22), arg3, (char **)&x3);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x3 = (long *)alloca((size_t)datasize);
    memset((void *)x3, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x3, _cffi_type(22), arg3) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(21), arg4, (char **)&x4);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x4 = (void * *)alloca((size_t)datasize);
    memset((void *)x4, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x4, _cffi_type(21), arg4) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(22), arg5, (char **)&x5);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x5 = (long *)alloca((size_t)datasize);
    memset((void *)x5, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x5, _cffi_type(22), arg5) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_GetRingBufferWriteRegions(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, long);
}
#else
#  define _cffi_f_PaUtil_GetRingBufferWriteRegions _cffi_d_PaUtil_GetRingBufferWriteRegions
#endif

static long _cffi_d_PaUtil_InitializeRingBuffer(PaUtilRingBuffer * x0, long x1, long x2, void * x3)
{
  return PaUtil_InitializeRingBuffer(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_InitializeRingBuffer(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  long x2;
  void * x3;
  Py_ssize_t datasize;
  long result;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "PaUtil_InitializeRingBuffer", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(9), arg0) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, long);
  if (x2 == (long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x3 = (void *)alloca((size_t)datasize);
    memset((void *)x3, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x3, _cffi_type(1), arg3) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_InitializeRingBuffer(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, long);
}
#else
#  define _cffi_f_PaUtil_InitializeRingBuffer _cffi_d_PaUtil_InitializeRingBuffer
#endif

static long _cffi_d_PaUtil_ReadRingBuffer(PaUtilRingBuffer * x0, void * x1, long x2)
{
  return PaUtil_ReadRingBuffer(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_ReadRingBuffer(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  void * x1;
  long x2;
  Py_ssize_t datasize;
  long result;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "PaUtil_ReadRingBuffer", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(9), arg0) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (void *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(1), arg1) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, long);
  if (x2 == (long)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_ReadRingBuffer(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, long);
}
#else
#  define _cffi_f_PaUtil_ReadRingBuffer _cffi_d_PaUtil_ReadRingBuffer
#endif

static long _cffi_d_PaUtil_WriteRingBuffer(PaUtilRingBuffer * x0, void const * x1, long x2)
{
  return PaUtil_WriteRingBuffer(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_WriteRingBuffer(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  void const * x1;
  long x2;
  Py_ssize_t datasize;
  long result;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "PaUtil_WriteRingBuffer", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (PaUtilRingBuffer *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(9), arg0) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(33), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (void const *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(33), arg1) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, long);
  if (x2 == (long)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_WriteRingBuffer(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, long);
}
#else
#  define _cffi_f_PaUtil_WriteRingBuffer _cffi_d_PaUtil_WriteRingBuffer
#endif

static int _cffi_d_callback(void * x0, void * x1, unsigned long x2, PaStreamCallbackTimeInfo const * x3, unsigned long x4, void * x5)
{
  return callback(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_callback(PyObject *self, PyObject *args)
{
  void * x0;
  void * x1;
  unsigned long x2;
  PaStreamCallbackTimeInfo const * x3;
  unsigned long x4;
  void * x5;
  Py_ssize_t datasize;
  int result;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "callback", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (void *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(1), arg0) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (void *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(1), arg1) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, unsigned long);
  if (x2 == (unsigned long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(4), arg3, (char **)&x3);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x3 = (PaStreamCallbackTimeInfo const *)alloca((size_t)datasize);
    memset((void *)x3, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x3, _cffi_type(4), arg3) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, unsigned long);
  if (x4 == (unsigned long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg5, (char **)&x5);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x5 = (void *)alloca((size_t)datasize);
    memset((void *)x5, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x5, _cffi_type(1), arg5) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = callback(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, int);
}
#else
#  define _cffi_f_callback _cffi_d_callback
#endif

static void _cffi_d_init_stream(Py_PaStream * x0)
{
  init_stream(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_init_stream(PyObject *self, PyObject *arg0)
{
  Py_PaStream * x0;
  Py_ssize_t datasize;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(43), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (Py_PaStream *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(43), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { init_stream(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_init_stream _cffi_d_init_stream
#endif

static void _cffi_d_reset_stream(Py_PaStream * x0)
{
  reset_stream(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_reset_stream(PyObject *self, PyObject *arg0)
{
  Py_PaStream * x0;
  Py_ssize_t datasize;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(43), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (Py_PaStream *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(43), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { reset_stream(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_reset_stream _cffi_d_reset_stream
#endif

static int _cffi_const_MAX_MESSAGE_LEN(unsigned long long *o)
{
  int n = (MAX_MESSAGE_LEN) <= 0;
  *o = (unsigned long long)((MAX_MESSAGE_LEN) | 0);  /* check that MAX_MESSAGE_LEN is an integer */
  if (!_cffi_check_int(*o, n, 128U))
    n |= 2;
  return n;
}

static int _cffi_const_paInputOverflow(unsigned long long *o)
{
  int n = (paInputOverflow) <= 0;
  *o = (unsigned long long)((paInputOverflow) | 0);  /* check that paInputOverflow is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_paInputUnderflow(unsigned long long *o)
{
  int n = (paInputUnderflow) <= 0;
  *o = (unsigned long long)((paInputUnderflow) | 0);  /* check that paInputUnderflow is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_paOutputOverflow(unsigned long long *o)
{
  int n = (paOutputOverflow) <= 0;
  *o = (unsigned long long)((paOutputOverflow) | 0);  /* check that paOutputOverflow is an integer */
  if (!_cffi_check_int(*o, n, 8U))
    n |= 2;
  return n;
}

static int _cffi_const_paOutputUnderflow(unsigned long long *o)
{
  int n = (paOutputUnderflow) <= 0;
  *o = (unsigned long long)((paOutputUnderflow) | 0);  /* check that paOutputUnderflow is an integer */
  if (!_cffi_check_int(*o, n, 4U))
    n |= 2;
  return n;
}

static int _cffi_const_paPrimingOutput(unsigned long long *o)
{
  int n = (paPrimingOutput) <= 0;
  *o = (unsigned long long)((paPrimingOutput) | 0);  /* check that paPrimingOutput is an integer */
  if (!_cffi_check_int(*o, n, 16U))
    n |= 2;
  return n;
}

_CFFI_UNUSED_FN
static void _cffi_checkfld__PaStreamCallbackTimeInfo(PaStreamCallbackTimeInfo *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  { double *tmp = &p->inputBufferAdcTime; (void)tmp; }
  { double *tmp = &p->currentTime; (void)tmp; }
  { double *tmp = &p->outputBufferDacTime; (void)tmp; }
}
struct _cffi_align__PaStreamCallbackTimeInfo { char x; PaStreamCallbackTimeInfo y; };

_CFFI_UNUSED_FN
static void _cffi_checkfld__PaUtilRingBuffer(PaUtilRingBuffer *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  (void)((p->bufferSize) | 0);  /* check that 'PaUtilRingBuffer.bufferSize' is an integer */
  (void)((p->writeIndex) | 0);  /* check that 'PaUtilRingBuffer.writeIndex' is an integer */
  (void)((p->readIndex) | 0);  /* check that 'PaUtilRingBuffer.readIndex' is an integer */
  (void)((p->bigMask) | 0);  /* check that 'PaUtilRingBuffer.bigMask' is an integer */
  (void)((p->smallMask) | 0);  /* check that 'PaUtilRingBuffer.smallMask' is an integer */
  (void)((p->elementSizeBytes) | 0);  /* check that 'PaUtilRingBuffer.elementSizeBytes' is an integer */
  { char * *tmp = &p->buffer; (void)tmp; }
}
struct _cffi_align__PaUtilRingBuffer { char x; PaUtilRingBuffer y; };

_CFFI_UNUSED_FN
static void _cffi_checkfld__Py_PaStream(Py_PaStream *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  (void)((p->status) | 0);  /* check that 'Py_PaStream.status' is an integer */
  { PaStreamCallbackTimeInfo *tmp = &p->lastTime; (void)tmp; }
  (void)((p->last_callback) | 0);  /* check that 'Py_PaStream.last_callback' is an integer */
  (void)((p->loop) | 0);  /* check that 'Py_PaStream.loop' is an integer */
  (void)((p->frame_count) | 0);  /* check that 'Py_PaStream.frame_count' is an integer */
  (void)((p->frames) | 0);  /* check that 'Py_PaStream.frames' is an integer */
  (void)((p->pad) | 0);  /* check that 'Py_PaStream.pad' is an integer */
  (void)((p->offset) | 0);  /* check that 'Py_PaStream.offset' is an integer */
  (void)((p->rxchannels) | 0);  /* check that 'Py_PaStream.rxchannels' is an integer */
  (void)((p->txchannels) | 0);  /* check that 'Py_PaStream.txchannels' is an integer */
  { unsigned char * *tmp = &p->rxmapping; (void)tmp; }
  { unsigned char * *tmp = &p->txmapping; (void)tmp; }
  { unsigned char * *tmp = &p->_mapping; (void)tmp; }
  (void)((p->xruns) | 0);  /* check that 'Py_PaStream.xruns' is an integer */
  (void)((p->txElementSize) | 0);  /* check that 'Py_PaStream.txElementSize' is an integer */
  (void)((p->rxElementSize) | 0);  /* check that 'Py_PaStream.rxElementSize' is an integer */
  (void)((p->inputOverflows) | 0);  /* check that 'Py_PaStream.inputOverflows' is an integer */
  (void)((p->inputUnderflows) | 0);  /* check that 'Py_PaStream.inputUnderflows' is an integer */
  (void)((p->outputOverflows) | 0);  /* check that 'Py_PaStream.outputOverflows' is an integer */
  (void)((p->outputUnderflows) | 0);  /* check that 'Py_PaStream.outputUnderflows' is an integer */
  { PaUtilRingBuffer * *tmp = &p->rxbuffer; (void)tmp; }
  { PaUtilRingBuffer * *tmp = &p->txbuffer; (void)tmp; }
  { char(*tmp)[128] = &p->errorMsg; (void)tmp; }
  (void)((p->_autoframes) | 0);  /* check that 'Py_PaStream._autoframes' is an integer */
}
struct _cffi_align__Py_PaStream { char x; Py_PaStream y; };

static const struct _cffi_global_s _cffi_globals[] = {
  { "MAX_MESSAGE_LEN", (void *)_cffi_const_MAX_MESSAGE_LEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "PaUtil_AdvanceRingBufferReadIndex", (void *)_cffi_f_PaUtil_AdvanceRingBufferReadIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 8), (void *)_cffi_d_PaUtil_AdvanceRingBufferReadIndex },
  { "PaUtil_AdvanceRingBufferWriteIndex", (void *)_cffi_f_PaUtil_AdvanceRingBufferWriteIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 8), (void *)_cffi_d_PaUtil_AdvanceRingBufferWriteIndex },
  { "PaUtil_FlushRingBuffer", (void *)_cffi_f_PaUtil_FlushRingBuffer, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 39), (void *)_cffi_d_PaUtil_FlushRingBuffer },
  { "PaUtil_GetRingBufferReadAvailable", (void *)_cffi_f_PaUtil_GetRingBufferReadAvailable, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 36), (void *)_cffi_d_PaUtil_GetRingBufferReadAvailable },
  { "PaUtil_GetRingBufferReadRegions", (void *)_cffi_f_PaUtil_GetRingBufferReadRegions, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 18), (void *)_cffi_d_PaUtil_GetRingBufferReadRegions },
  { "PaUtil_GetRingBufferWriteAvailable", (void *)_cffi_f_PaUtil_GetRingBufferWriteAvailable, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 36), (void *)_cffi_d_PaUtil_GetRingBufferWriteAvailable },
  { "PaUtil_GetRingBufferWriteRegions", (void *)_cffi_f_PaUtil_GetRingBufferWriteRegions, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 18), (void *)_cffi_d_PaUtil_GetRingBufferWriteRegions },
  { "PaUtil_InitializeRingBuffer", (void *)_cffi_f_PaUtil_InitializeRingBuffer, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 12), (void *)_cffi_d_PaUtil_InitializeRingBuffer },
  { "PaUtil_ReadRingBuffer", (void *)_cffi_f_PaUtil_ReadRingBuffer, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 26), (void *)_cffi_d_PaUtil_ReadRingBuffer },
  { "PaUtil_WriteRingBuffer", (void *)_cffi_f_PaUtil_WriteRingBuffer, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 31), (void *)_cffi_d_PaUtil_WriteRingBuffer },
  { "Py_PaStream_default", (void *)_cffi_const_Py_PaStream_default, _CFFI_OP(_CFFI_OP_CONSTANT, 47), (void *)0 },
  { "callback", (void *)_cffi_f_callback, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_callback },
  { "init_stream", (void *)_cffi_f_init_stream, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 42), (void *)_cffi_d_init_stream },
  { "paInputOverflow", (void *)_cffi_const_paInputOverflow, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "paInputUnderflow", (void *)_cffi_const_paInputUnderflow, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "paOutputOverflow", (void *)_cffi_const_paOutputOverflow, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "paOutputUnderflow", (void *)_cffi_const_paOutputUnderflow, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "paPrimingOutput", (void *)_cffi_const_paPrimingOutput, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "reset_stream", (void *)_cffi_f_reset_stream, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 42), (void *)_cffi_d_reset_stream },
};

static const struct _cffi_field_s _cffi_fields[] = {
  { "inputBufferAdcTime", offsetof(PaStreamCallbackTimeInfo, inputBufferAdcTime),
                          sizeof(((PaStreamCallbackTimeInfo *)0)->inputBufferAdcTime),
                          _CFFI_OP(_CFFI_OP_NOOP, 52) },
  { "currentTime", offsetof(PaStreamCallbackTimeInfo, currentTime),
                   sizeof(((PaStreamCallbackTimeInfo *)0)->currentTime),
                   _CFFI_OP(_CFFI_OP_NOOP, 52) },
  { "outputBufferDacTime", offsetof(PaStreamCallbackTimeInfo, outputBufferDacTime),
                           sizeof(((PaStreamCallbackTimeInfo *)0)->outputBufferDacTime),
                           _CFFI_OP(_CFFI_OP_NOOP, 52) },
  { "bufferSize", offsetof(PaUtilRingBuffer, bufferSize),
                  sizeof(((PaUtilRingBuffer *)0)->bufferSize),
                  _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "writeIndex", offsetof(PaUtilRingBuffer, writeIndex),
                  sizeof(((PaUtilRingBuffer *)0)->writeIndex),
                  _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "readIndex", offsetof(PaUtilRingBuffer, readIndex),
                 sizeof(((PaUtilRingBuffer *)0)->readIndex),
                 _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "bigMask", offsetof(PaUtilRingBuffer, bigMask),
               sizeof(((PaUtilRingBuffer *)0)->bigMask),
               _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "smallMask", offsetof(PaUtilRingBuffer, smallMask),
                 sizeof(((PaUtilRingBuffer *)0)->smallMask),
                 _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "elementSizeBytes", offsetof(PaUtilRingBuffer, elementSizeBytes),
                        sizeof(((PaUtilRingBuffer *)0)->elementSizeBytes),
                        _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "buffer", offsetof(PaUtilRingBuffer, buffer),
              sizeof(((PaUtilRingBuffer *)0)->buffer),
              _CFFI_OP(_CFFI_OP_NOOP, 48) },
  { "status", offsetof(Py_PaStream, status),
              sizeof(((Py_PaStream *)0)->status),
              _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "lastTime", offsetof(Py_PaStream, lastTime),
                sizeof(((Py_PaStream *)0)->lastTime),
                _CFFI_OP(_CFFI_OP_NOOP, 45) },
  { "last_callback", offsetof(Py_PaStream, last_callback),
                     sizeof(((Py_PaStream *)0)->last_callback),
                     _CFFI_OP(_CFFI_OP_NOOP, 53) },
  { "loop", offsetof(Py_PaStream, loop),
            sizeof(((Py_PaStream *)0)->loop),
            _CFFI_OP(_CFFI_OP_NOOP, 56) },
  { "frame_count", offsetof(Py_PaStream, frame_count),
                   sizeof(((Py_PaStream *)0)->frame_count),
                   _CFFI_OP(_CFFI_OP_NOOP, 57) },
  { "frames", offsetof(Py_PaStream, frames),
              sizeof(((Py_PaStream *)0)->frames),
              _CFFI_OP(_CFFI_OP_NOOP, 54) },
  { "pad", offsetof(Py_PaStream, pad),
           sizeof(((Py_PaStream *)0)->pad),
           _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "offset", offsetof(Py_PaStream, offset),
              sizeof(((Py_PaStream *)0)->offset),
              _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "rxchannels", offsetof(Py_PaStream, rxchannels),
                  sizeof(((Py_PaStream *)0)->rxchannels),
                  _CFFI_OP(_CFFI_OP_NOOP, 56) },
  { "txchannels", offsetof(Py_PaStream, txchannels),
                  sizeof(((Py_PaStream *)0)->txchannels),
                  _CFFI_OP(_CFFI_OP_NOOP, 56) },
  { "rxmapping", offsetof(Py_PaStream, rxmapping),
                 sizeof(((Py_PaStream *)0)->rxmapping),
                 _CFFI_OP(_CFFI_OP_NOOP, 55) },
  { "txmapping", offsetof(Py_PaStream, txmapping),
                 sizeof(((Py_PaStream *)0)->txmapping),
                 _CFFI_OP(_CFFI_OP_NOOP, 55) },
  { "_mapping", offsetof(Py_PaStream, _mapping),
                sizeof(((Py_PaStream *)0)->_mapping),
                _CFFI_OP(_CFFI_OP_NOOP, 55) },
  { "xruns", offsetof(Py_PaStream, xruns),
             sizeof(((Py_PaStream *)0)->xruns),
             _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "txElementSize", offsetof(Py_PaStream, txElementSize),
                     sizeof(((Py_PaStream *)0)->txElementSize),
                     _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "rxElementSize", offsetof(Py_PaStream, rxElementSize),
                     sizeof(((Py_PaStream *)0)->rxElementSize),
                     _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "inputOverflows", offsetof(Py_PaStream, inputOverflows),
                      sizeof(((Py_PaStream *)0)->inputOverflows),
                      _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "inputUnderflows", offsetof(Py_PaStream, inputUnderflows),
                       sizeof(((Py_PaStream *)0)->inputUnderflows),
                       _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "outputOverflows", offsetof(Py_PaStream, outputOverflows),
                       sizeof(((Py_PaStream *)0)->outputOverflows),
                       _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "outputUnderflows", offsetof(Py_PaStream, outputUnderflows),
                        sizeof(((Py_PaStream *)0)->outputUnderflows),
                        _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "rxbuffer", offsetof(Py_PaStream, rxbuffer),
                sizeof(((Py_PaStream *)0)->rxbuffer),
                _CFFI_OP(_CFFI_OP_NOOP, 9) },
  { "txbuffer", offsetof(Py_PaStream, txbuffer),
                sizeof(((Py_PaStream *)0)->txbuffer),
                _CFFI_OP(_CFFI_OP_NOOP, 9) },
  { "errorMsg", offsetof(Py_PaStream, errorMsg),
                sizeof(((Py_PaStream *)0)->errorMsg),
                _CFFI_OP(_CFFI_OP_NOOP, 50) },
  { "_autoframes", offsetof(Py_PaStream, _autoframes),
                   sizeof(((Py_PaStream *)0)->_autoframes),
                   _CFFI_OP(_CFFI_OP_NOOP, 56) },
};

static const struct _cffi_struct_union_s _cffi_struct_unions[] = {
  { "PaStreamCallbackTimeInfo", 45, _CFFI_F_CHECK_FIELDS,
    sizeof(PaStreamCallbackTimeInfo), offsetof(struct _cffi_align__PaStreamCallbackTimeInfo, y), 0, 3 },
  { "PaUtilRingBuffer", 46, _CFFI_F_CHECK_FIELDS,
    sizeof(PaUtilRingBuffer), offsetof(struct _cffi_align__PaUtilRingBuffer, y), 3, 7 },
  { "Py_PaStream", 47, _CFFI_F_CHECK_FIELDS,
    sizeof(Py_PaStream), offsetof(struct _cffi_align__Py_PaStream, y), 10, 24 },
};

static const struct _cffi_typename_s _cffi_typenames[] = {
  { "PaStreamCallbackFlags", 3 },
  { "PaStreamCallbackTimeInfo", 45 },
  { "PaTime", 52 },
  { "PaUtilRingBuffer", 46 },
  { "Py_PaStream", 47 },
  { "ring_buffer_size_t", 10 },
};

static const struct _cffi_type_context_s _cffi_type_context = {
  _cffi_types,
  _cffi_globals,
  _cffi_fields,
  _cffi_struct_unions,
  NULL,  /* no enums */
  _cffi_typenames,
  20,  /* num_globals */
  3,  /* num_struct_unions */
  0,  /* num_enums */
  6,  /* num_typenames */
  NULL,  /* no includes */
  59,  /* num_types */
  0,  /* flags */
};

#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit__py_pastream(const void *p[])
{
    p[0] = (const void *)0x2601;
    p[1] = &_cffi_type_context;
#if PY_MAJOR_VERSION >= 3
    return NULL;
#endif
}
#  ifdef _MSC_VER
     PyMODINIT_FUNC
#  if PY_MAJOR_VERSION >= 3
     PyInit__py_pastream(void) { return NULL; }
#  else
     init_py_pastream(void) { }
#  endif
#  endif
#elif PY_MAJOR_VERSION >= 3
PyMODINIT_FUNC
PyInit__py_pastream(void)
{
  return _cffi_init("_py_pastream", 0x2601, &_cffi_type_context);
}
#else
PyMODINIT_FUNC
init_py_pastream(void)
{
  _cffi_init("_py_pastream", 0x2601, &_cffi_type_context);
}
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif
